/* +-------------+---------+-----------------------+
   | John Breaux | jab0910 | JohnBreaux@my.unt.edu |
   +-------------+---------+-----------------------+
   |              Created  2022-04-17              |
   +-----------------------------------------------+  */
#include <fstream>
#include <regex>
#include "graph.hpp"

//* Vector of regex rules, in order of precedence. Earlier entries will be checked before later entries.
std::vector<std::regex> patterns = {
   //* Rule 1: A line containing % is a comment, and should not be parsed
   std::regex("(%)"),
   //* Rule 2: A line containing `num_processes` contains the number of processes
   std::regex("(num_processes)\\D*(\\d+)"),
   //* Rule 3: A line containing `num_resources` contains the number of resources
   std::regex("(num_resources)\\D*(\\d+)"),
   //* Rule 4: A line containing integers should be saved somewhere
   std::regex("(-?[0-9]+.*-?[0-9]+)") // Magical!
};

// Say hi to Stovey, here to bake your cakes.
//* convert string s to vector<int> no matter the cost
std::vector<int> stovi (const std::string &s) {
   // Create the number classification rule
   std::regex rule("-?[0-9]+");
   // Make a new tokenizing iterator using the rule
   std::sregex_token_iterator start{s.begin(), s.end(), rule, 0}, end;
   // Match the numbers
   std::vector<int> res;
   for (auto i = start; i != end; ++i) res.push_back(atoi(i->str().c_str()));
   // give the numbers back
   return res;
}

//* Read from a file of name `filename`
int graph::read(std::string filename) {
   // Open file with name filename as input fstream
   std::ifstream f(filename);
   // Read the entire file
   while (!f.eof()) {
      // acquire a line
      std::string line;
      std::getline (f, line);
      // Iterate over each pattern, and grab the associated data
      for (const auto &pattern: patterns) {
         std::smatch res;
         // Capture data with a regex search, using the pattern
         if (std::regex_search (line, res, pattern)) {
            // get the pattern type, value
            std::string type = res.format("$1"), value = res.format("$2");
            // Handle the pattern
            //   Rule 1: If line is a comment, ignore it and move on
            if (type == "%") break;

            //   Rule 2: If num_processes= matched, assign value to num_processes
            if (type == "num_processes") { num_processes = std::stoi(value); } else

            //   Rule 3: If num_resources= matched, assign value to num_resources
            if (type == "num_resources") { num_resources = std::stoi(value); } else

            //   Rule 4: If this line is a comma-separated list of numbers:
            //     If resource_counts is empty, assume this is the resource counts,
            //     else it's part of the matrix being reloaded
            if (!resource_counts.size()) { resource_counts = stovi(type);    } else
            { m.push_back(stovi(type)); }
            break;
         }
      }
   }

   //* Test validity of the input file
   // Check if all sections are properly defined:
   const char errmsg[] = "%s: Definition failed for %s\n";
   int ret = 0;
   if (!num_processes){
      printf(errmsg, filename.c_str(), "num_processes");
      ret |= 0b00001;
   }
   if (!num_resources){
      printf(errmsg, filename.c_str(), "num_resources");
      ret |= 0b00010;
   }
   if (resource_counts.size() != num_resources){
      printf(errmsg, filename.c_str(), "resource counts");
      ret |= 0b00100;
   }
   if (m.size() != num_processes + num_resources){
      printf(errmsg, filename.c_str(), "adjacency matrix rows");
      ret |= 0b01000;
   }
   for (auto i: m) {
      if (i.size() != num_processes + num_resources)
         ret |= 0b10000;
   }
   if (ret & 0b10000) {
      printf(errmsg, filename.c_str(), "adjacency matrix columns");
   }
   if (ret) {
      printf("Check to ensure your input file is correct and follows the Rules.\n");
   }

   // fstream destructor closes file when going out of scope
   // in this house we RAII sometimes.
   return ret;
}